# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Workday API client.

As the Python Workday client is no longer maintained this client has been created to
allow collection of log data.
"""

import logging
import time
from typing import Any, Dict, Optional

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://{base_url}/ccx/api/privacy/v1/{identity}"
API_PAGE_SIZE = 100


class Client:
    def __init__(
        self,
        base_url: Optional[str] = None,
        identity: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        refresh_token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new client.

        :param base_url: The Workday instance base url
        :param identity: The Workday tenant name
        :param client_id: The Workday integration client id
        :param client_secret: The Workday integration client secret
        :param refresh_token: The allocated Workday integration client refresh token
        :param retry: Whether to automatically retry if recoverable errors are
            encountered, such as rate-limiting.
        """
        self.base_url = base_url
        self.identity = identity
        self.client_secret = client_secret
        self.client_id = client_id
        self.refresh_token = refresh_token
        self.retry = retry
        self.logger = logging.getLogger(__name__)

        self.headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self._api_base_uri = API_BASE_URI.format(base_url=base_url, identity=identity)

    def _get(
        self, url: str, params: Optional[Dict[str, Optional[str]]] = None
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: A URL to perform the HTTP GET against.
        :param parameters: An optional set of HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # Retry on rate-limit, but only if requested.
                if getattr(err.response, "status_code", None) == 429:
                    self.logger.warning("Rate-limit was exceeded during request")
                    if self.retry:
                        time.sleep(int(err.response.headers.get("Retry-After", "1")))
                        continue
                    else:
                        raise RateLimitException(err)

                raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def _post(
        self,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Optional[str]]] = None,
    ) -> Dict[str, Any]:
        """A POST wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP POST against.
        :param headers: Dictionary of headers to add to the request.
        :param data: HTTP parameters to add to the request.

        :raises RequestFailedException: An HTTP request failed.

        :return: The response to the request in JSON.
        """
        try:
            response = requests.post(
                url,
                headers=headers,
                data=data,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise RequestFailedException(err)

        return response.json()

    def get_access_token(self):
        """Exchange a refresh token for an access token.

        This is required by Workday to auth the integration and then grant the bearer
        token to access the API.

        :return: If the request is successful, the bearer token is returned to
            the Client class header.
        """
        url = f"https://{self.base_url}/ccx/oauth2/{self.identity}/token"

        bearer_response = self._post(
            url,
            data={
                "client_id": f"{self.client_id}",
                "client_secret": f"{self.client_secret}",
                "grant_type": "refresh_token",
                "refresh_token": f"{self.refresh_token}",
            },
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )

        access_token = bearer_response.get("access_token")
        self.headers["Authorization"] = f"Bearer {access_token}"

    def get_activity_logging(
        self,
        cursor: int = 0,
        to_date: Optional[str] = None,
        from_date: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit logs which match the provided filters.

        :param from_date: The required date and time of the earliest log entry.
        :param to_date: The required date and time in UTC of the latest log entry.
        :param limit: The maximum number of items to include in a single response.
        :param cursor: The index position of the first object in a response collection.
            Cursor to use when fetching events (pagination).

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        # The endpoint returns the same total value of results regardless of the limit
        # and offset parameters. The pagination parameters determine the amount of
        # content in the data[] array.
        result = self._get(
            f"{self._api_base_uri}/activityLogging",
            params={
                "from": from_date,
                "to": to_date,
                "instancesReturned": "3",
                "limit": str(API_PAGE_SIZE),
                "offset": str(cursor),
            },
        )

        data = result.body.get("data", [])

        # keep paging until we meet the total number of results
        if len(data) == API_PAGE_SIZE:
            cursor += API_PAGE_SIZE
        else:
            cursor = 0

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=data)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Google GSuite Activities connector for Grove."""

import json
from datetime import datetime, timedelta

import google_auth_httplib2
import httplib2
from google.auth.exceptions import GoogleAuthError
from google.oauth2 import service_account
from googleapiclient.discovery import build
from googleapiclient.errors import Error as GACError

from grove.connectors import BaseConnector
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import (
    ConfigurationException,
    NotFoundException,
    RequestFailedException,
)

# This connector is only interested in the GSuite Reports API.
SCOPES = ["https://www.googleapis.com/auth/admin.reports.audit.readonly"]


def as_rfc3339(date: datetime) -> str:
    """Formats an input date into RFC3339 format.

    :param date: The input datetime object to convert to RFC3339 format.

    :return: An RFC 3339 format date as a string.
    """
    return date.isoformat(sep="T", timespec="milliseconds") + "Z"


class Connector(BaseConnector):
    NAME = "gsuite_activities"
    POINTER_PATH = "id.time"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    @property
    def delay(self):
        """Defines the amount of time to delay collection of logs (in minutes).

        This is used to allow time for logs to become 'consistent' before they are
        collected. Google backfill log entries based on their published lag guidelines.
        As a result of this, collection of events within this lag window may result in
        missed events.

        As a result of these constraints, this value is configurable to allow operators
        to preference consistency over speed of delivery, and vice versa. For example,
        a delay of 20 would instruct Grove to only collect logs after they are at least
        20 minutes old.

        This defaults to 0 (no delay).

        :return: The "delay" component of the connector configuration.
        """
        try:
            candidate = self.configuration.delay
        except AttributeError:
            return 0

        try:
            candidate = int(candidate)
        except ValueError as err:
            raise ConfigurationException(
                f"Configured 'delay' is not valid. Value must be an integer. {err}"
            )

        return candidate

    def collect(self):
        """Collects activities from the Google GSuite Reports API.

        As the Google APIs use OAuth 2.0 2LO ('two-legged OAuth') which contains a
        number of fields inside of a JSON 'service account file' the key and identity
        are treated a little differently in this connector.

        Rather than the key being a single authentication token, the key should contain
        the entire 'service account file' in JSON format - as generated by the Google
        API console.

        The identity must be the name of a service account which has been granted domain
        wide delegation. Please see the following guides for more information:

          https://developers.google.com/admin-sdk/directory/v1/guides/delegation
          https://developers.google.com/admin-sdk/reports/reference/rest/v1/activities

        :raises RequestFailedException: An HTTP request failed.
        """
        cursor = str()
        http = google_auth_httplib2.AuthorizedHttp(
            self.get_credentials(),
            http=self.get_http_transport(),
        )

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the GSuite reports API the pointer is
        # the value of the id.time field from the latest record retrieved from the API.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = as_rfc3339(datetime.utcnow() - timedelta(days=7))

        # Page over all activities since the last pointer.
        more_requests = True

        with build("admin", "reports_v1", http=http) as service:
            while more_requests:
                # Transform the dates back to native to allow timedelta and comparison.
                start = datetime.fromisoformat(self.pointer.replace("Z", ""))
                end = datetime.utcnow() - timedelta(minutes=self.delay)

                if end <= start:
                    self.logger.debug(
                        "Collection end time is prior to start, skipping.",
                        extra={
                            "start": as_rfc3339(start),
                            "end": as_rfc3339(end),
                            **self.log_context,
                        },
                    )
                    break

                if cursor:
                    self.logger.debug(
                        "Using pageToken as cursor to request next page of activities.",
                        extra={"cursor": cursor, **self.log_context},
                    )
                    request = service.activities().list(
                        userKey="all",
                        applicationName=self.operation,
                        startTime=as_rfc3339(start),
                        endTime=as_rfc3339(end),
                        pageToken=cursor,
                    )
                else:
                    self.logger.debug(
                        "Using startTime from pointer to request activity list.",
                        extra={
                            "start": as_rfc3339(start),
                            "end": as_rfc3339(end),
                            **self.log_context,
                        },
                    )
                    request = service.activities().list(
                        userKey="all",
                        startTime=as_rfc3339(start),
                        endTime=as_rfc3339(end),
                        applicationName=self.operation,
                    )

                # As both the _entries and activities objects are Lists, we can extend
                # them in order to support appending more data on any subsequent
                # iterations (due to paging).
                try:
                    results = request.execute()
                    activities = results.get("items", [])

                    self.logger.debug(
                        "Got activities from the GSuite API.",
                        extra={
                            "count": len(activities),
                            **self.log_context,
                        },
                    )
                    self.save(activities)
                except (GoogleAuthError, GACError) as err:
                    raise RequestFailedException(f"Request to GSuite API failed: {err}")

                # Determine whether we're still paging.
                if "nextPageToken" not in results:
                    self.logger.debug(
                        "No nextPageToken, finishing collection.",
                        extra=self.log_context,
                    )
                    more_requests = False
                    break

                self.logger.debug(
                    "nextPageToken is present, paging.", extra=self.log_context
                )
                cursor = results["nextPageToken"]
                more_requests = True

    def get_http_transport(self):
        """Creates an HTTP object for use by the Google API Client.

        :return: An httplib2.Http object for use with the Google API client.
        """
        return httplib2.Http()

    def get_credentials(self):
        """Generates and returns a credentials instance from the connector's configured
        service account info. This is used for required to perform operations using the
        Google API client.

        :return: A credentials instance built from configured service account info.

        :raises ConfigurationException: There is an issue with the configuration
            for this connector.
        """
        try:
            service_account_info = json.loads(self.key)
        except json.JSONDecodeError as err:
            raise ConfigurationException(
                f"Unable to load service account JSON for {self.identity}: {err}"
            )

        # Construct the credentials, including scopes and delegation.
        try:
            credentials = service_account.Credentials.from_service_account_info(
                service_account_info,
                scopes=SCOPES,
                subject=self.identity,
            )
        except ValueError as err:
            raise ConfigurationException(
                "Unable to generate credentials from service account info for "
                f" {self.identity}: {err}"
            )

        return credentials



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""1Password Audit event log connector for Grove."""

import datetime

from grove.connectors import BaseConnector
from grove.connectors.onepassword.api import Client
from grove.constants import CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "onepassword_events_audit"
    POINTER_PATH = "timestamp"
    LOG_ORDER = CHRONOLOGICAL

    def collect(self):
        """Collects all logs from the 1Password audit event log API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(token=self.key)
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the 1Password API this is an ISO
        # timestamp in a field called "timestamp".
        try:
            _ = self.pointer
        except NotFoundException:
            week_ago = datetime.datetime.now() - datetime.timedelta(days=7)
            self.pointer = (week_ago).astimezone().replace(microsecond=0).isoformat()

        # Get log data from the upstream API, paging as required.
        while True:
            log = client.get_auditevents(start_time=self.pointer, cursor=cursor)

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements integration tests for the GSuite Alerts collector."""

import os
import unittest
from unittest.mock import patch

from googleapiclient.http import HttpMockSequence

from grove.connectors.gsuite.alerts import Connector
from grove.models import ConnectorConfig
from tests import mocks


class MockSequence(HttpMockSequence):
    def close(self, *args, **kwargs):
        pass


class GSuiteAlertsTestCase(unittest.TestCase):
    """Implements integration tests for the GSuite Alerts collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="1FEEDFEED1",
                key="{}",
                name="test",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @patch("google.oauth2.service_account.Credentials.from_service_account_info")
    @patch("googleapiclient.discovery.build")
    @patch("grove.connectors.gsuite.alerts.Connector.get_http_transport")
    def test_collect_pagination(self, mock_transport, mock_request, mock_auth):
        """Ensure collection works as expected."""
        mock_transport.return_value = MockSequence(
            [
                (
                    {"status": "200"},
                    open(
                        os.path.join(self.dir, "fixtures/gsuite/alerts/001.json"),
                        "r",
                    ).read(),
                ),
                (
                    {"status": "200"},
                    open(
                        os.path.join(self.dir, "fixtures/gsuite/alerts/002.json"),
                        "r",
                    ).read(),
                ),
            ],
        )
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 2)
        self.assertEqual(self.connector.pointer, "2021-04-03T14:05:39.950458Z")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides a consistent logger in Grove."""


import json
import logging
from typing import Any, Dict

from aws_lambda_powertools.logging.formatter import RESERVED_LOG_ATTRS, JsonFormatter


class GroveFormatter(JsonFormatter):
    """A logging formatter which emits logs in a consistent and structured way.

    This formatter emits logs in JSON, with all "context" provided at creation added to
    each emitted message, alongside any "extra" data provided during logging calls. This
    formatter also adds the function name, file name, and line number of the log call to
    each message.
    """

    def __init__(self, context: Dict[str, str], *args, **kwargs):
        self.utc = True
        self.context = context

        super().__init__(*args, **kwargs)

        # Add the function name the code to the code, including line-number, to all
        # messages to enable easier location of the source of any issues.
        self.reserved_attrs = (*RESERVED_LOG_ATTRS, "function")

        self.log_format["location"] = "%(pathname)s:%(lineno)d"
        self.log_format["function"] = "%(funcName)s"

    def extract_keys(self, record: logging.LogRecord) -> Dict[str, Any]:
        """Extracts and formats log records into dictionaries ready for serialisation.

        This is heavily based on the code from the AWS Lambda PowerTools formatter that
        this formatter inherits from.

        :param record: A log record to process.

        :return: A dictionary of log data to be serialized and output.
        """
        extras = {}
        structured = record.__dict__.copy()
        structured["asctime"] = self.formatTime(record=record)

        # Determine which arguments are "extras" provided by the caller.
        for key, value in structured.items():
            if key not in self.reserved_attrs:
                extras[key] = value

        # Format all fields according to their specified format strings.
        formatted = {}

        for key, value in self.log_format.items():
            if value and key in self.reserved_attrs:
                formatted[key] = value % structured
            else:
                formatted[key] = value

        # Stuff all 'extras' provided to a logging call into a field called "details".
        formatted.update({"detail": extras})

        return formatted

    def format(self, record: logging.LogRecord) -> str:  # noqa: A003
        """Return the log message including any context provided by the entrypoint.

        :param record: A log record to process.

        :return: A stringified JSON document rendered from the log record.
        """
        structured = {}

        # Remove any records which have a value of None.
        candidate = self.extract_keys(record=record)
        candidate["message"] = str(record.msg)
        candidate["context"] = self.context

        for key, value in candidate.items():
            if value is not None:
                structured[key] = value

        return json.dumps(structured, default=str)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements e2e tests for grove using the heartbeat connector."""

import json
import os
import subprocess
import unittest


class GroveCLITestCase(unittest.TestCase):
    """Implements e2e tests for grove."""

    def test_grove_cli(self):
        """Ensure messages are created and printed to stdout."""
        connectors_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__), "../templates/deployment/local-quick-start"
            )
        )

        env = os.environ | {
            "GROVE_OUTPUT_HANDLER": "local_stdout",
            "GROVE_CONFIG_HANDLER": "local_file",
            "GROVE_CACHE_HANDLER": "local_memory",
            "GROVE_CONFIG_LOCAL_FILE_PATH": connectors_path,
        }

        out = subprocess.run(["grove"], capture_output=True, env=env)
        self.assertEqual(out.returncode, 0)

        for logline in out.stderr.split(b"\n"):
            if logline:
                self.assertIn(json.loads(logline)["level"], ["INFO", "DEBUG"])

        parsed_loglines = []
        for logline in out.stdout.split(b"\n"):
            if logline:
                parsed_loglines.append(json.loads(logline))

        self.assertEqual(len(parsed_loglines), 5)
        self.assertEqual(parsed_loglines[0]["message"]["type"], "heartbeat")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove stdout output handler."""

import datetime
import json
from typing import Any, Dict, List, Optional

from grove.constants import DATESTAMP_FORMAT, GROVE_METADATA_KEY
from grove.exceptions import DataFormatException
from grove.outputs import BaseOutput


class Handler(BaseOutput):
    def submit(
        self,
        data: bytes,
        connector: str,
        identity: str,
        operation: str,
        part: int = 0,
        kind: Optional[str] = "json",
        descriptor: Optional[str] = "raw",
    ):
        """Print captured data to stdout.

        :param data: Log data to write.
        :param connector: Name of the connector which retrieved the data.
        :param identity: Identity the collected data was collect for.
        :param operation: Operation the collected logs are associated with.
        :param part: Number indicating which part of the same log stream this file
            contains data for. This is used to indicate that the logs are from the same
            collection, but have been broken into smaller files for downstream
            processing.
        :param kind: The format of the data being output.
        :param descriptor: An arbitrary descriptor to identify the data being output.

        """
        datestamp = datetime.datetime.utcnow()

        for entry in data.decode("utf-8").splitlines():
            print(
                json.dumps(
                    {
                        "part": part,
                        "kind": kind,
                        "descriptor": descriptor,
                        "connector": connector,
                        "identity": identity,
                        "operation": operation,
                        "datestamp": datestamp.strftime(DATESTAMP_FORMAT),
                        "message": json.loads(entry),
                    }
                ),
                flush=True,
            )

    def serialize(self, data: List[Any], metadata: Dict[str, Any] = {}) -> bytes:
        """Serialize data to a standard format (NDJSON).

        :param data: A list of log entries to serialize to JSON.
        :param metadata: Metadata to append to each log entry before serialization. If
            not specified no metadata will be added.

        :return: Log data serialized as NDJSON.

        :raises DataFormatException: Cannot serialize the input to JSON.
        """
        candidate = []

        for entry in data:
            entry[GROVE_METADATA_KEY] = metadata

            # We don't want to silently drop and lose single records, so drop the entire
            # batch if there is bad data (which will trigger a retry next run).
            try:
                candidate.append(json.dumps(entry, separators=(",", ":")))
            except TypeError as err:
                message = f"Unable to serialize to JSON: {err}"
                raise DataFormatException(message)

        return bytes("\r\n".join(candidate), "utf-8")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides functions used between entrypoints."""

import os
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict, List

from aws_lambda_powertools import Logger

from grove.constants import (
    DEFAULT_CONFIG_HANDLER,
    DEFAULT_WORKER_COUNT,
    ENV_GROVE_CONFIG_HANDLER,
    ENV_GROVE_SECRET_HANDLER,
    ENV_GROVE_WORKER_COUNT,
    PLUGIN_GROUP_CONFIG,
    PLUGIN_GROUP_CONNECTOR,
    PLUGIN_GROUP_SECRET,
)
from grove.exceptions import GroveException
from grove.helpers import plugin
from grove.logging import GroveFormatter
from grove.models import ConnectorConfig


def dispatch(config: ConnectorConfig, context: Dict[str, str]):
    """Executes a connector, blocking until complete.

    This function is intended to be called via a ThreadPoolExecutor to enable concurrent
    execution of connectors.

    :param config: A connector configuration object for this connector thread.
    :param context: Contextual information relating to the current runtime.
    """
    handler = plugin.lookup_handler(config.connector, PLUGIN_GROUP_CONNECTOR).load()
    instance = handler(config, context)
    instance.run()


def configure() -> List[ConnectorConfig]:
    """Fetches all configuration documents and associated secrets."""
    configs = plugin.load_handler(
        os.environ.get(ENV_GROVE_CONFIG_HANDLER, DEFAULT_CONFIG_HANDLER),
        PLUGIN_GROUP_CONFIG,
    )

    # Immediately ignore configuration documents if they're marked as disabled.
    loaded = []

    for configuration in configs.get():
        if configuration.disabled:
            continue

        loaded.append(configuration)

    # Secret backends are optional, so if there isn't one defined, assume secrets are
    # embedded in the configuration.
    handler = os.environ.get(ENV_GROVE_SECRET_HANDLER)

    if not handler:
        return loaded

    # Load all required secrets into the configuration objects from the configured
    # secrets backend.
    secrets = plugin.load_handler(handler, PLUGIN_GROUP_SECRET)
    return secrets.load(loaded)


def entrypoint(context: Dict[str, Any]):
    """Provides the main entrypoint for Grove.

    This function should be called from various wrappers in order to execute Grove when
    running under the respective runtime. This is in order to enable use in serverless,
    containerised, virtualised, and "bare metal" environments.

    :param context: Contextual information relating to the current runtime.
    """
    logger = Logger(
        "grove",
        logger_formatter=GroveFormatter(context),
        stream=sys.stderr,
    )
    logger.info("Grove started")

    # Attempt to load connector configuration, failure to do so is not recoverable.
    try:
        configurations = configure()
    except GroveException as err:
        logger.critical(
            "Failed to initialise configuration handler", extra={"exception": err}
        )
        return

    # Allow users to optionally configure the number of worker threads.
    try:
        workers = int(os.environ.get(ENV_GROVE_WORKER_COUNT, DEFAULT_WORKER_COUNT))
    except ValueError as err:
        logger.critical(
            f"Worker count ('{ENV_GROVE_WORKER_COUNT}') must be a number.",
            extra={"exception": err},
        )
        return

    # All connectors will be executed in their own thread, up to the maximum configured
    # workers specified by the worker count.
    logger.info("Spawning thread pool for connectors", extra={"workers": workers})

    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = {}

        for configuration in configurations:
            future = pool.submit(dispatch, configuration, context)
            futures[future] = configuration

        # Blocks until all threads have exited.
        for future in as_completed(futures):
            configuration = futures[future]
            del futures[future]

            # Last ditch effort to catch any unhandled exceptions to ensure that they're
            # logged out.
            try:
                future.result()
            except GroveException as err:
                logger.error(
                    "Connector exited abnormally.",
                    extra={
                        "exception": err,
                        "configuration": configuration.name,
                        "connector": configuration.connector,
                    },
                )

            # Exit logs are recorded on error / exception and success.
            logger.info(
                "Connector has exited.",
                extra={
                    "configuration": configuration.name,
                    "connector": configuration.connector,
                },
            )

    logger.info("All connectors have exited. Grove execution has finished.")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""SalesForce Event Log connector for Grove."""

import csv
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List

import requests
from simple_salesforce import Salesforce, SalesforceLogin
from simple_salesforce.exceptions import SalesforceError

from grove.connectors import BaseConnector
from grove.constants import CHRONOLOGICAL
from grove.exceptions import (
    ConfigurationException,
    NotFoundException,
    RequestFailedException,
)

# TODO: Make this dynamic?
SF_VERSION = "51.0"
SF_OPERATIONS = ["Login"]
SF_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"

# SOQL query templates for use when accessing logs.
SOQL_EVENTLOGFILE = (
    "SELECT Id, ApiVersion, EventType, CreatedDate, LogDate, LogFile "
    "FROM EventLogFile "
    "WHERE EventType = '{event}' "
    "AND LogDate >= {pointer}"
)


class Connector(BaseConnector):
    NAME = "sf_event_log"
    POINTER_PATH = "TIMESTAMP_DERIVED"
    LOG_ORDER = CHRONOLOGICAL

    @property
    def token(self):
        """Fetches the SalesForce token from the configuration.

        This is required as this is a third authentication element required by SF.

        :return: The "token" portion of the connector's configuration.
        """
        try:
            return self.configuration.token
        except AttributeError:
            return None

    def collect(self):  # noqa: C901
        """Collects EventLogs from the SF Cloud API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.

        :raises RequestFailedException: An HTTP request failed.
        :raises ConfigurationException: An issue was found with the configuration for
            this connector.
        """
        session = requests.session()
        try:
            (sf_session, sf_instance) = SalesforceLogin(
                session=session,
                username=self.identity,
                password=self.key,
                security_token=self.token,
            )
            client = Salesforce(
                instance=sf_instance,
                session_id=sf_session,
                version=SF_VERSION,
            )
        except (SalesforceError, requests.exceptions.RequestException) as err:
            raise RequestFailedException(
                f"Unable to authenticate with SalesForce: {err}"
            )

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago.
        now = datetime.now(timezone.utc)

        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (now - timedelta(days=7)).strftime(SF_TIMESTAMP_FORMAT)

        # Pointers are stored as strings, so cast to a datetime object for use when
        # constructing filters later on.
        pointer_native = datetime.strptime(self.pointer, SF_TIMESTAMP_FORMAT)

        if self.operation not in SF_OPERATIONS:
            raise ConfigurationException(
                f"Operation must be one of {SF_OPERATIONS}, got '{self.operation}'"
            )

        log_files: Dict[str, List[Any]] = {}
        next_records_url = None

        while True:
            # Page if required. Otherwise, fetch ALL log entries from midnight on the
            # day the last pointer was recorded at. This is required as the SalesForce
            # API appears to split EventLogFiles by date, and does not allow filtering
            # inside of logs using SOQL. See below for more information.
            try:
                if next_records_url is not None:
                    records = client.query_more(next_records_url)  # type: ignore
                else:
                    records = client.query_all(
                        SOQL_EVENTLOGFILE.format(
                            event=self.operation,
                            pointer=pointer_native.strftime("%Y-%m-%dT00:00:00.00Z"),
                        ),
                    )
            except (SalesforceError, requests.exceptions.RequestException) as err:
                raise RequestFailedException(
                    f"Unable to query SalesForce for event logs: {err}"
                )

            # Fetch a list of logs, which will be fetched and processed later - as the
            # SalesForce API returns a REFERENCE to log files here, not the log data
            # itself.
            for record in records.get("records", []):
                record_type = record.get("EventType")
                if record_type not in log_files:
                    log_files[record_type] = []

                log_files[self.operation].append(
                    {
                        "Id": record.get("Id"),
                        "LogFile": record.get("LogFile"),
                        "ApiVersion": record.get("ApiVersion"),
                        "LogDate": record.get("LogDate"),
                        "CreatedDate": record.get("CreatedDate"),
                    }
                )

            # Determine if more requests are needed, otherwise, break out.
            if records.get("nextRecordsUrl") is None:
                break

            next_records_url = records.get("nextRecordsUrl")

        # Manually process out events after the pointer. This is required as the SOQL
        # WHERE filter doesn't appear to allow searching on LogFile contents, only on
        # the LogDate / CreatedDate metadata, which isn't helpful.
        for log_type in log_files:
            for log_file in log_files.get(log_type, {}):
                # Use Requests to get the LogFile directly, passing in the session id
                # from our current session. It doesn't appear that Simple Salesforce
                # has a nice way to handle this for us.
                try:
                    request = requests.get(
                        f"https://{client.sf_instance}/{log_file.get('LogFile')}",
                        headers={
                            "Content-Type": "application/json",
                            "Authorization": f"Bearer {client.session_id}",
                        },
                    )
                except requests.exceptions.RequestException as err:
                    raise RequestFailedException(
                        f"Unable to retrieve event log from SalesForce: {err}"
                    )

                # Convert the CSV (?!) into a nicely JSON serialisable format.
                entries = []

                for entry in csv.DictReader(str(request.content, "utf-8").split("\n")):
                    # Skip if the entry is BEFORE the known pointer - this is required
                    # to handle partial logs from SalesForce. This is expensive, but it
                    # should reduce the need for deduplication later in the pipeline.
                    entry_time = datetime.strptime(
                        entry["TIMESTAMP_DERIVED"], SF_TIMESTAMP_FORMAT
                    )
                    if entry_time.timestamp() <= pointer_native.timestamp():
                        continue

                    entries.append(entry)

                self.save(entries)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the Torq API Client."""

import base64
import os
import re
import unittest
from unittest.mock import patch

import responses
from responses import matchers

from grove.connectors.torq.api import Client
from tests import mocks


class TorqClientTestCases(unittest.TestCase):
    """Implements unit tests for the Torq API Client."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.identity = "CLIENT_ID"
        self.key = "CLIENT_SECRET"
        self.client = None

    @responses.activate
    def test_get_bearer_token(self):
        """Ensure we get bearer token from client-id client-secret."""
        set_expected_basic_auth = str(
            base64.b64encode(bytes(f"{self.identity}:{self.key}", "utf-8")),
            "utf-8",
        )
        # auth response
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/torq/client/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
            match=[
                matchers.header_matcher(
                    {"Authorization": f"Basic {set_expected_basic_auth}"}
                )
            ],
        )
        # token hardcoded from fixtures file
        resp = {"access_token": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}
        self.assertIsNone(self.client)
        self.client = Client(identity=self.identity, key=self.key)
        self.assertEqual(resp.get("access_token"), self.client.bearer_token)

        # make sure the authorization header is set with the new bearer token
        self.assertEqual(
            f"Bearer {resp.get('access_token')}",
            self.client.headers.get("Authorization"),
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove AWS Lambda entrypoint."""

from typing import Any, Dict

from aws_lambda_powertools.utilities.typing import LambdaContext

from grove.entrypoints import base


def entrypoint(event: Dict[str, Any], context: LambdaContext) -> Dict[str, Any]:
    """Grove AWS Lambda wrapper.

    :param event: Unused event data for this AWS Lambda invocation.
    :param context: Execution context from AWS Lambda.

    :return: An unused status message.
    """
    base.entrypoint(
        context={
            "runtime": __file__,
            "runtime_id": context.aws_request_id,
            "lambda_function_arn": context.invoked_function_arn,
            "lambda_function_memory_size": context.memory_limit_in_mb,
            "lambda_request_id": context.aws_request_id,
        },
    )
    return {"Status": "OKAY"}


# Support local development if called as a script.
if __name__ == "__main__":
    # This breaks encapsulation, but the PowerTools provided helpers don't expose a way
    # to set properties - as they're designed to just be for type hints. We're not going
    # to extend the class only for testing, so here we are.
    context = LambdaContext()
    context._aws_request_id = "C0FFEEC0-FFEE-C0FF-EEC0-FFEEC0FFEEC0"
    context._memory_limit_in_mb = 128
    context._invoked_function_arn = "arn:aws:lambda:local:0123456789012:grove"

    entrypoint({}, context)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the S3 output handler."""

import os
import unittest

import boto3
from moto import mock_s3

from grove.exceptions import ConfigurationException
from grove.outputs.aws_s3 import Handler


class S3OutputTestCase(unittest.TestCase):
    """Implements tests for the S3 output handler."""

    @mock_s3
    def setUp(self):
        self.client = boto3.resource("s3", region_name="us-east-1")
        self.client.create_bucket(Bucket="PersistentBucket")

        os.environ["GROVE_OUTPUT_AWS_S3_BUCKET"] = "PersistentBucket"

    @mock_s3
    def test_bucket_check(self):
        """Ensures the bucket location constraint check operates properly."""
        # Ensure the handler is setup when a properly configured bucket is specified.
        Handler()

        # Ensure that an exception is raised when no bucket is specified.
        if "GROVE_OUTPUT_AWS_S3_BUCKET" in os.environ:
            del os.environ["GROVE_OUTPUT_AWS_S3_BUCKET"]

        with self.assertRaises(ConfigurationException):
            Handler()



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Terraform Cloud audit trail API client.

The official TFC SDK does not currently support the Audit API, so interactions need
to be handled manually.
"""

import logging
import time
from typing import Dict, Optional

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://app.terraform.io/api/v2"


class Client:
    def __init__(
        self,
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ) -> None:
        """Setup a new client.

        :param token: The TFC API Bearer token.
        :param retry: Whether to automatically retry if recoverable errors are
            encountered, such as rate-limiting.
        """
        self.retry = retry
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {token}",
        }

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # Retry on rate-limit, but only if requested.
                if getattr(err.response, "status_code", None) == 429:
                    self.logger.warning("Rate-limit was exceeded during request")
                    if self.retry:
                        time.sleep(1)
                        continue
                    else:
                        raise RateLimitException(err)

                raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_trails(
        self,
        since: Optional[str] = None,
        cursor: Optional[int] = 1,
        page_size: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit logs which match the provided filters.

        :param since: The ISO8601 date of the most recent event to include (inclusive).
        :param cursor: The page to fetch. If omitted, endpoint returns first page.
        :param page_size: Number of audit events per page. Defaults to 1000.

        :return: AuditLogEntries object containing a pagination cursor, and log
            entries.
        """
        # See psf/requests issue #2651 for why we can happily pass in None values
        # and not have the request key added to the URI.
        result = self._get(
            f"{API_BASE_URI}/organization/audit-trail",
            params={
                "since": since,
                "page[number]": str(cursor),
                "page[size]": page_size,
            },
        )
        cursor = result.body.get("pagination", {}).get("next_page", 0)

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body.get("data", []))



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Okta API client."""

import datetime
import logging
import time
from typing import Dict, Optional
from urllib.parse import unquote

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://{identity}.{domain}"
API_PAGE_SIZE = 1000


class Client:
    def __init__(
        self,
        domain: str = "okta.com",
        identity: Optional[str] = None,
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new client.

        :param identity: The Okta subdomain.
        :param domain: The Okta domain, without customer subdomain.
        :param token: The Okta API key.
        :param retry: Whether to automatically retry if recoverable errors are
            encountered, such as rate-limiting.
        """
        self.retry = retry
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"SSWS {token}",
        }

        self._api_base_uri = API_BASE_URI.format(identity=identity, domain=domain)

    def _parse_link_header(self, link: str) -> str:
        """Attempts to parse a URL from the provided Link header.

        :param link: The value of a Link header returned from a previous request.

        :raises ValueError: No, or an invalid, Link header was encountered.

        :return: A clean and ready to use URL from the Link header provided by Okta.
        """
        # A link header may contain N entries ("next" and "self").
        url = None
        next_url = None
        self_url = None
        links = link.split(",")

        for entry in links:
            link_parts = entry.split(";")
            link_rel = link_parts[-1].strip()

            if "next" in link_rel:
                next_url = unquote(link_parts[0].strip().lstrip("<").rstrip(">"))

            elif "self" in link_rel:
                self_url = unquote(link_parts[0].strip().lstrip("<").rstrip(">"))

        # If self_url and next_url are equal than this is the last page, return a
        # value error, otherwise return the next page url.
        if not next_url or next_url == self_url:
            raise ValueError()

        # Try to mitigate SSRFs where a baked Link header is returned.
        url = next_url

        if self._api_base_uri.lower() not in url.lower():
            raise ValueError(
                f"{self._api_base_uri} not found in Link header ({url}). Ignoring."
            )

        return url

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle pagination for the caller.

        :param url: A URL to perform the HTTP GET against.
        :param parameters: An optional set of HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # 429s are used with a header to indicate rate-limit exceeded.
                if getattr(err.response, "status_code", None) != 429:
                    raise RequestFailedException(err)

                if err.response.headers.get("X-RateLimit-Remaining") != "0":
                    raise RequestFailedException(err)

                # Retry on rate-limit, but only if requested.
                self.logger.warning("Rate-limit was exceeded during request")
                if self.retry:
                    time_current = int(
                        datetime.datetime.now(datetime.timezone.utc).strftime("%s")
                    )
                    time_ratelimit_reset = int(
                        err.response.headers.get("X-Rate-Limit-Reset", time_current)
                    )

                    # If there was no X-Rate-Limit-Reset header, or the time is in
                    # the past, just wait for a second before retry.
                    if time_ratelimit_reset > time_ratelimit_reset:
                        time.sleep(time_ratelimit_reset - time_current)
                    else:
                        time.sleep(1)

                    continue
                else:
                    raise RateLimitException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_audit_logs(
        self,
        since: Optional[str] = None,
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Get log data from the upstream API.

        The Okta system log API uses Polling Requests and Bounded Requests and has a
        request limit of 1000 per minute:

            https://developer.okta.com/docs/reference/api/system-log/#request-parameters

        :param since: Filters the upper time bound of the log events published
            property for bounded queries or persistence time for polling queries.
        :param cursor: Cursor to use when fetching results. Supersedes other
            parameters.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        # Use the cursor URL if set, otherwise construct the initial query.
        if cursor is not None:
            self.logger.debug(
                "Collecting next page using Okta provided cursor",
                extra={"cursor": cursor},
            )
            result = self._get(cursor)
        else:
            result = self._get(
                f"{self._api_base_uri}/api/v1/logs",
                params={
                    "since": since,
                    "sortOrder": "ASCENDING",
                    "limit": str(API_PAGE_SIZE),
                },
            )

        # Track the results.
        try:
            cursor = self._parse_link_header(result.headers.get("Link", ""))
        except ValueError:
            cursor = None

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body)  # type: ignore



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Slack Audit connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.slack.api import Client
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "slack_audit_logs"
    POINTER_PATH = "date_create"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    def collect(self):
        """Collects all logs from the Slack Audit API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(token=self.key)
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the Slack audit API the pointer is the
        # value of the "date_create" field from the latest record retrieved from the
        # API - which is in seconds since epoch ("Unix Time") format.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.utcnow() - timedelta(days=7)).strftime("%s")

        # Page over data using the cursor, saving returned data page by page.
        while True:
            log = client.get_logs(oldest=self.pointer, cursor=cursor)

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Custom types used throughout Grove."""

from collections.abc import MutableMapping
from typing import Any, Dict, List, NamedTuple, Optional, Union


class HTTPResponse(NamedTuple):
    """Provides both the headers and the body of an HTTP response."""

    headers: MutableMapping  # type: ignore
    body: Dict[str, Any]


class AuditLogEntries(NamedTuple):
    """Provides both a pagination cursor and entries from an API response."""

    cursor: Union[Optional[str], Optional[int]]
    entries: List[Any]



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""SalesForce Marketing Cloud audit API client.

The FuelSDK for Python is quite a heavy client - as it handles both SOAP and REST. The
REST interface is not far off simply using requests either. As a result, to keep things
lightweight this client simply implements the required operations for Get Security
Events and Get Audit Events. There does not appear to be any special handling in this
client - such as for rate-limits or authentication.

This may be replaced in future with the FuelSDK, but for now, we'll keep it simple.
"""

import logging
from typing import Dict, Optional

import requests

from grove.exceptions import RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://{identity}.rest.marketingcloudapis.com"
API_PAGE_SIZE = 500


class Client:
    def __init__(
        self,
        identity: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """Setup a new client.

        Args:
            identity: The SFMC subdomain.
            token: The SFMC API Bearer token.
        """
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {token}",
        }

        # We need to push the identity into the URI, so we'll keep track of this.
        self._api_base_uri = API_BASE_URI.format(identity=identity)

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        try:
            response = requests.get(url, headers=self.headers, params=params)
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json()[0])

    def get_audit(
        self,
        kind: str,
        cursor: int = 1,
        startdate: Optional[str] = None,
        enddate: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit events from the relevant audit REST API.

        :param kind: The kind of audit events to fetch.
        :param cursor: The page number of the records to fetch.
        :param latest: The ISO8601 timestamp of the most recent event (inclusive).
        :param oldest: The ISO8601 timestamp of the least recent event (inclusive).

        :return: AuditLogEntries object containing a pagination cursor, and log
            entries.
        """
        # See psf/requests issue #2651 for why we can happily pass in None values
        # and not have the request key added to the URI.
        result = self._get(
            f"{self._api_base_uri}/data/v1/audit/{kind}",
            params={
                "startdate": startdate,
                "enddate": enddate,
                "$pagesize": str(API_PAGE_SIZE),
                "$page": str(cursor),
            },
        )

        # Keep paging until we run out of results.
        if result.body.get("count") == API_PAGE_SIZE:
            cursor += 1
        else:
            cursor = 0

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body.get("items", []))

    def get_audit_events(
        self,
        cursor: int = 1,
        startdate: Optional[str] = None,
        enddate: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit events from the getAuditEvents REST API.

        :param cursor: The page number of the records to fetch.
        :param latest: The ISO8601 timestamp of the most recent event (inclusive).
        :param oldest: The ISO8601 timestamp of the least recent event (inclusive).

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_audit(
            kind="auditEvents",
            cursor=cursor,
            startdate=startdate,
            enddate=enddate,
        )

    def get_security_events(
        self,
        cursor: int = 1,
        startdate: Optional[str] = None,
        enddate: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of security events from the getSecurityEvents REST API.

        :param cursor: The page number of the records to fetch.
        :param latest: The ISO8601 timestamp of the most recent event (inclusive).
        :param oldest: The ISO8601 timestamp of the least recent event (inclusive).

        :return: A list of security events.
        """
        return self.get_audit(
            kind="securityEvents",
            cursor=cursor,
            startdate=startdate,
            enddate=enddate,
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements integration tests for the Workday Audit collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses

from grove.connectors.workday.activity_logging import Connector
from grove.exceptions import RequestFailedException
from grove.models import ConnectorConfig
from tests import mocks


class WorkdayActivityLogging(unittest.TestCase):
    """Implements integration tests for the Workday Audit collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                base_url="test.workday.com",
                client_id="000000",
                client_secret="EEEEE",
                identity="corporation",
                key="token",
                name="test",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_client_rate_limit(self):
        """Ensure ratelimit waiting is working as expected."""
        # Get bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/001.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        # Rate limit the first request.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=429,
            content_type="application/json",
            body=bytes(),
            headers={
                "X-Rate-Limit-Remaining": "0",
                "X-Rate-Limit-Reset": "0",  # Unix-time, so definitely in the past :)
            },
        )
        # Succeed on the second.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/004.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Ensure time.sleep is called with the correct value in response to a
        # rate-limit.
        with patch("time.sleep", return_value=None) as mock_sleep:
            self.connector.collect()
            mock_sleep.assert_called_with(1)

    @responses.activate
    def test_get_token_fail_on_server_error(self):
        """Ensure server errors raise an appropriate exception."""
        # Setup an fake HTTP 400 response from WorkDay.
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=400,
            content_type="application/json",
            body=bytes(),
        )

        with self.assertRaises(RequestFailedException):
            self.connector.collect()

    @responses.activate
    def test_collect_fail_on_server_error(self):
        """Ensure server errors raise an appropriate exception."""
        # Get bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/001.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        # Setup an fake HTTP 500 response from WorkDay.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=500,
            body=bytes(),
        )

        with self.assertRaises(RequestFailedException):
            self.connector.collect()

    @responses.activate
    def test_collect_pagination(self):
        """Ensure pagination is working as expected."""
        # Get bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/001.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        # Succeed with a cursor returned (to indicate paging is required).
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/002.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # The last "page" returns an empty cursor.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/004.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Check the pointer matches the latest execution_time value, and that the
        # expected number of logs were returned.
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 113)
        self.assertEqual(self.connector.pointer, "2021-10-12T23:50:09.752Z")

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        # Get bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/001.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/004.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Ensure only a single value is returned, and the pointer is properly set.
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 13)
        self.assertEqual(self.connector.pointer, "2021-10-12T23:50:09.752Z")

    @responses.activate
    def test_collect_no_results(self):
        """Ensure break accurs when there is no results returned."""
        # Get bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/001.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/workday/activity_logging/003.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 0)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Atlassian Audit connector for Grove."""

from datetime import datetime, timedelta, timezone

from grove.connectors import BaseConnector
from grove.connectors.atlassian.api import Client
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException

ISO_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"
ISO_TIMESTAMP_FORMAT_NO_MS = "%Y-%m-%dT%H:%M:%S%z"


class Connector(BaseConnector):
    NAME = "atlassian_audit_events"
    POINTER_PATH = "attributes.time"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    def collect(self):
        """Collects all logs from the Atlassian Events API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(identity=self.identity, token=self.key)
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the Atlassian events API the pointer is
        # the value of the "time" field from the latest record retrieved from the
        # API - which is an ISO8601 timestamp.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.now(timezone.utc) - timedelta(days=7)).strftime(
                ISO_TIMESTAMP_FORMAT
            )

        # Atlassians org API from_date is The earliest date and time of the event
        # represented as a UNIX epoch time. To support this, the pointers need to be
        # CONVERTED to UNIX epoch time up to the millisecond from the value of the
        # pointer.
        #
        # The try / except block is required as two different time formats have been
        # seen returned from Atlassian for these event types: one with milliseconds, one
        # without.
        try:
            start = int(
                datetime.strptime(self.pointer, ISO_TIMESTAMP_FORMAT).timestamp() * 1000
            )
        except ValueError:
            start = int(
                datetime.strptime(self.pointer, ISO_TIMESTAMP_FORMAT_NO_MS).timestamp()
                * 1000
            )

        # Page over data using the cursor, saving returned data page by page.
        while True:
            log = client.get_audit(from_date=str(start), cursor=cursor)

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove AWS SSM parameter store configuration handler."""

import json
import logging
import os
from typing import List, Optional

from boto3.session import Session
from botocore.exceptions import BotoCoreError, ClientError
from pydantic import BaseSettings, Field, ValidationError

from grove.configs import BaseConfig
from grove.exceptions import AccessException, ConfigurationException
from grove.helpers import parsing
from grove.models import ConnectorConfig


class Configuration(BaseSettings):
    """Defines environment variables used to configure the AWS SSM handler.

    This should also include any appropriate default values for fields which are not
    required.
    """

    prefix: Optional[str] = Field(
        default="/grove/connectors/",
        description="A prefix to added to the beginning of all parameter store paths.",
    )
    assume_role_arn: Optional[str] = Field(
        description="An optional AWS role to assume when authenticating with AWS.",
        default=None,
    )
    ssm_region: Optional[str] = Field(
        description="The region that the parameter store exists in (default us-east-1)",
        default=os.environ.get("AWS_REGION", "us-east-1"),
    )

    class Config:
        """Allow environment variable override of configuration fields.

        This also enforce a prefix for all environment variables for this handler. As
        an example the field `assume_role_arn` would be set using the environment
        variable `GROVE_CONFIG_AWS_SSM_ASSUME_ROLE_ARN`.
        """

        env_prefix = "GROVE_CONFIG_AWS_SSM_"
        case_insensitive = True


class Handler(BaseConfig):
    """A configuration handler to read configuration documents from AWS SSM."""

    def __init__(self):
        """Sets up access to AWS SSM.

        This handler also attempt to assume a configured role in order to allow
        cross-account use - if required.

        :raises ConfigurationException: There was an issue with configuration.
        :raises AccessException: An issue occurred when accessing SSM.
        """
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = Configuration()
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))

        # Explicit calls to session are mostly used to allow mocks during testing.
        session = Session()

        try:
            if not self.config.assume_role_arn:
                self._ssm = session.client("ssm", region_name=self.config.ssm_region)
            else:
                sts = session.client("sts")
                self.logger.debug(
                    "Config handler is attempting to assume AWS role for SSM",
                    extra={"role_arn": self.config.assume_role_arn},
                )

                role = sts.assume_role(
                    RoleArn=self.config.assume_role_arn,
                    RoleSessionName="GroveConfigurationHandler",
                )
                self._ssm = session.client(
                    "ssm",
                    region_name=self.config.ssm_region,
                    aws_access_key_id=role["Credentials"]["AccessKeyId"],
                    aws_secret_access_key=role["Credentials"]["SecretAccessKey"],
                    aws_session_token=role["Credentials"]["SessionToken"],
                )
        except (ClientError, BotoCoreError, KeyError) as err:
            raise AccessException(f"Config handler failed to access AWS SSM: {err}")

    def get(self, id: str = "/") -> List[ConnectorConfig]:
        """Gets and returns one or more connector configuration objects from AWS SSM.

        Configuration documents are enumerated by their path, allowing multiple
        connectors to be defined in the SSM parameter store under a common path.

        :param id: The path of the parameter(s) to return. This is combined with the
            configured prefix, if set.

        :raises AccessException: An issue occurred when querying the value from SSM.

        :return: A list of connector configuration objects.
        """
        # Always combine the configured prefix with the specified path.
        path = f"{self.config.prefix.rstrip('/')}/{id.lstrip('/')}"  # type: ignore

        # Handling is performed in two-stages to reduce nested exception handlers and
        # to prevent a single bad connector configuration from causing all to fail.
        # Failure to load a single configuration should NEVER cause Grove to fail.
        pager = self._ssm.get_paginator("get_parameters_by_path")
        pages = pager.paginate(Path=path, Recursive=True, WithDecryption=True)
        candidates = {}

        try:
            for page in pages:
                for parameter in page.get("Parameters", {}):
                    name = parameter.get("Name")
                    value = parameter.get("Value")

                    if name and value:
                        candidates[name] = value
                        continue

                    self.logger.error(
                        "Config handler failed to fetch a connector configuration",
                        extra={"path": path, "key": name},
                    )
        except (ClientError, BotoCoreError) as err:
            raise AccessException(
                f"Config handler failed to read value from SSM path '{path}': {err}"
            )

        # Generate a list of documents for later processing.
        connectors = []

        for name, value in candidates.items():
            try:
                connectors.append(ConnectorConfig(**json.loads(value)))
            except (json.JSONDecodeError, ValidationError) as err:
                self.logger.error(
                    "Unable to load connector configuration",
                    extra={"document": name, "exception": err},
                )
                continue

        return connectors



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Constants used throughout Grove."""

# The prefix for all pointers in the cache.
CACHE_KEY_LOCK = "execution_lock"
CACHE_KEY_SEEN = "deduplication"
CACHE_KEY_POINTER = "pointer"
CACHE_KEY_POINTER_NEXT = "pointer_next"
CACHE_KEY_POINTER_PREV = "pointer_previous"

# The prefix for window start pointers.
CACHE_KEY_WINDOW_START = "window_start"
CACHE_KEY_WINDOW_END = "window_end"

# The default operation name to use where none is specified.
OPERATION_DEFAULT = "all"

# The common datestamp format to use for all date operations.
DATESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
LOCK_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

# Constants for tracking ordering of source log data.
CHRONOLOGICAL = "CHRONOLOGICAL"
REVERSE_CHRONOLOGICAL = "REVERSE_CHRONOLOGICAL"

# Specify the key under which grove metadata will be appended to a log entry.
GROVE_METADATA_KEY = "_grove"

# Environment variable names, used to override runtime settings.
ENV_GROVE_CACHE_HANDLER = "GROVE_CACHE_HANDLER"
ENV_GROVE_OUTPUT_HANDLER = "GROVE_OUTPUT_HANDLER"
ENV_GROVE_CONFIG_HANDLER = "GROVE_CONFIG_HANDLER"
ENV_GROVE_SECRET_HANDLER = "GROVE_SECRET_HANDLER"  # noqa: S105
ENV_GROVE_TELEMETRY_URI = "GROVE_TELEMETRY_URI"
ENV_GROVE_WORKER_COUNT = "GROVE_WORKER_COUNT"
ENV_GROVE_LOCK_DURATION = "GROVE_LOCK_DURATION"

# Plugin groups (setuptools entrypoints).
PLUGIN_GROUP_CACHE = "grove.caches"
PLUGIN_GROUP_OUTPUT = "grove.outputs"
PLUGIN_GROUP_CONFIG = "grove.configs"
PLUGIN_GROUP_PROCESSOR = "grove.processors"
PLUGIN_GROUP_SECRET = "grove.secrets"  # noqa: S105
PLUGIN_GROUP_CONNECTOR = "grove.connectors"

# Define defines for unset environment variables.
DEFAULT_CACHE_HANDLER = "local_memory"
DEFAULT_OUTPUT_HANDLER = "local_stdout"
DEFAULT_CONFIG_HANDLER = "local_file"

# Maximum number of connectors to execute concurrently.
DEFAULT_WORKER_COUNT = 50
DEFAULT_LOCK_DURATION = 300  # seconds.



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove HashiCorp Vault secret handler."""

import logging
import urllib.parse
from typing import Optional, Tuple

import jmespath
import requests
from pydantic import BaseSettings, Field, ValidationError

from grove.exceptions import AccessException, ConfigurationException
from grove.helpers import parsing
from grove.secrets import BaseSecret


class Configuration(BaseSettings):
    """Defines environment variables used to configure the HashiCorp Vault handler.

    This should also include any appropriate default values for fields which are not
    required.
    """

    addr: str = Field(
        description="The address of the Vault instance to retrieve secrets from.",
    )
    token: Optional[str] = Field(
        description="An optional vault token to use when authenticating with Vault.",
        default=None,
    )
    token_file: Optional[str] = Field(
        description="An optional file to read the Vault token from.",
        default=None,
    )
    namespace: Optional[str] = Field(
        description="An optional Vault namespace that should be used.",
        default=None,
    )
    api_version: str = Field(
        description="An optional Vault API version to use (default: v1).",
        default="v1",
    )

    class Config:
        """Allow environment variable override of configuration fields.

        This also enforce a prefix for all environment variables for this handler. As
        an example the field `token` would be set using the environment variable
        `GROVE_SECRET_HASHICORP_VAULT_TOKEN`.
        """

        env_prefix = "GROVE_SECRET_HASHICORP_VAULT_"
        case_insensitive = True


class Handler(BaseSecret):
    def __init__(self):
        """Sets up access to Vault.

        This backend performs a pre-flight to validate that the configured token is
        able to be used to query vault.

        :raises ConfigurationException: There was an issue with configuration.
        :raises AccessException: An issue occurred attempting to access Vault.
        """
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = Configuration()  # type: ignore
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))

        # If a token file is set any value passed as a token will be overwritten with
        # the value from file.
        if self.config.token_file:
            try:
                with open(self.config.token_file, "r") as fin:
                    self.config.token = fin.readline().strip()
            except OSError as err:
                raise ConfigurationException(
                    "Secrets handler could not read Vault token the configured token "
                    f"file of {self.config.token_file}: {err}"
                )

        # 'None' values will be automatically removed removed by requests when an HTTP
        # call is performed.
        self._headers = {
            "X-Vault-Token": self.config.token,
            "X-Vault-Request": "true",
            "X-Vault-Namespace": self.config.namespace,
        }

        # Perform a quick pre-flight to validate whether the credentials are valid.
        self._url = "/".join(
            [
                self.config.addr.rstrip("/"),
                self.config.api_version,
            ]
        )

        try:
            response = requests.get(
                f"{self._url}/auth/token/lookup-self",
                headers=self._headers,  # type: ignore
                allow_redirects=False,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise AccessException(
                f"Secrets handler could not access Vault at {self.config.addr}: {err}"
            )

    def get_field_and_path(self, path: str) -> Tuple[str, str]:
        """Extracts and removes 'field' parameters from a provided secret path.

        :param path: The path from the connector configuration to process.

        :raises VaultError: An error occurred while parsing data from the path.

        :returns: A tuple containing an extracted field, if any, and a Vault API
            compatible path.
        """
        url = urllib.parse.urlparse(path)
        qs = urllib.parse.parse_qs(url.query)

        # Extract and remove the field from the query parameters - if present.
        try:
            field = qs.pop("field", [])[0]
        except IndexError:
            raise ValueError("No 'field' parameter was found in the secret path.")

        # Regenerate the URL without the removed parameter, removing other parameters
        # we do not want configurable via the path.
        url = url._replace(netloc="", scheme="", params="")
        url = url._replace(query=urllib.parse.urlencode(qs, doseq=True))

        return field, urllib.parse.urlunparse(url).lstrip("/")

    def get(self, id: str) -> str:
        """Gets and returns a secret from Vault.

        To allow accessing different values under a configured secret path, this method
        uses a non-standard convention to encode which "field" of a returned credential
        is desired. This mimics the behavior of the Vault CLI "-field" option - though
        this is not a supported HTTP parameter by the Vault API directly.

        As an example of this, the following path would provide access to the 'password'
        portion of a credential stored in a KVv2 engine mounted at 'secret/':

            secret/data/example/demo?field=password

        To instead access a 'token' portion of a credential stored in the same path, the
        following would be used:

            secret/data/example/demo?field=token

        Finally, to perform the same operation against a KVv1 engine mounted at 'kv/'
        the path is almost the same. However, the '/data/' must ALSO be dropped, as this
        is only required for KVv2:

            kv/example/demo?field=token

        :param id: The path of the secret to retrieve - including engine.
        :param name: The name of the secret, defined by the connector configuration. If
            a 'field' is specified in the secret path this parameter will be ignored.

        :raises AccessException: An issue occurred when getting the secret from Vault.

        :returns: The plain-text secret from vault.
        """
        try:
            field, path = self.get_field_and_path(id)
        except ValueError as err:
            raise AccessException(
                f"Secrets handler could parse the provided Vault path of {id}. {err}"
            )

        try:
            response = requests.get(
                f"{self._url}/{path}",
                headers=self._headers,  # type: ignore
                allow_redirects=False,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise AccessException(
                f"Secrets handler failed to read secret from Vault path {id}. {err}"
            )

        # Return the first string value which is located using the desired name. This is
        # intended to return the first result without consideration of the engine.
        paths = [f"data.{field}", f"data.data.{field}"]
        secrets = response.json()

        for candidate in paths:
            secret = jmespath.search(candidate, secrets)
            if isinstance(secret, str):
                return secret

        raise AccessException(
            f"Secrets handler could not get field {field} from Vault path {path}"
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Zoom activity log connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.zoom import api
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException

DATESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class Connector(BaseConnector):
    NAME = "zoom_activities"
    POINTER_PATH = "time"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    @property
    def client_id(self):
        """Fetches the Zoom ClientID from the configuration.

        This is required as this is a third authentication element required by Zoom.
        """
        try:
            return self.configuration.client_id
        except AttributeError:
            return None

    def collect(self):
        """Collects all logs from the Zoom Activity API."""
        client = api.Client(
            identity=self.identity,
            client_id=self.client_id,
            key=self.key,
        )

        # Use the refresh token to generate a temporary access token.
        client.get_access_token()

        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a day ago. This API uses a YYYYMMDD date range. We can input using
        # RFC 3339 format because the API converts the to and from to YYYYMMDD.
        now = datetime.utcnow().strftime(DATESTAMP_FORMAT)
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.utcnow() - timedelta(days=1)).strftime(
                DATESTAMP_FORMAT
            )

        # Get log data from the upstream API. "From" and "to" datetime query parameters
        # are required. This API only gets data from a YYYYMMDD date range.
        while True:
            log = client.get_activities(
                from_date=self.pointer, to_date=now, cursor=cursor
            )

            # Zoom is troublesome as their API only allows filtering by date, not time.
            # As a result of this we need to do an additional step to ignore anything
            # after our pointer.
            candidates = self.deduplicate_by_pointer(log.entries)

            # Save this batch of log entries.
            self.save(log.entries)

            # If any log entries were filtered, the pointer was found, so we can skip
            # everything else.
            if len(candidates) != len(log.entries):
                break

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



"""Implements an example set of tests."""

import unittest


class ExampleTestCase(unittest.TestCase):
    """Implements an example set of tests."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        pass

    def tearDown(self):
        """Ensure everything is torn down between tests."""
        pass

    def test_example(self):
        """Always passes."""
        self.assertEqual(True, True)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the base output handler."""

import gzip
import unittest

from grove.outputs import BaseOutput


# Required as BaseOutput is an ABC, so without defining submit we will not be able to
# instantiate it to validate methods on the base class.
class TestOutput(BaseOutput):
    __test__ = False

    def submit(self, *args, **kwargs):
        pass


class BaseOutputTestCase(unittest.TestCase):
    """Implements tests for the base output handler."""

    def test_serialize(self):
        """Ensures serialization into NDJSON is functional."""
        handler = TestOutput()

        candidate = [
            {"id": "0001", "name": "One"},
            {"id": "0002", "name": "Two"},
            {"id": "0003", "name": "Three"},
        ]

        # Manually constructed.
        expected_raw = "\r\n".join(
            [
                '{"id":"0001","name":"One","_grove":{"field":"value"}}',
                '{"id":"0002","name":"Two","_grove":{"field":"value"}}',
                '{"id":"0003","name":"Three","_grove":{"field":"value"}}',
            ]
        )
        expected_gzip = gzip.compress(bytes(expected_raw, "utf-8"))

        self.assertEqual(
            expected_gzip,
            handler.serialize(data=candidate, metadata={"field": "value"}),
        )
